#ifndef HejHepMCInterface_H
#define HejHepMCInterface_H

#include <fstream>
#include <iostream>
#include "HepMC/PythiaWrapper.h"
#include "HepMC/IO_HEPEVT.h"
#include "HepMC/GenEvent.h"


class HejHepMCInterface {

public:

  HejHepMCInterface() {}

  HejHepMCInterface(std::string filename) {
    open(filename);
  }

  ~HejHepMCInterface() {
    close();
  }

  void open(std::string filename) {
    ofs.open(filename.c_str());
    ofs.precision(16);
    HepMC::write_HepMC_IO_block_begin(ofs);
  }

  void close() {
    HepMC::write_HepMC_IO_block_end(ofs);
    ofs.close();
  }

  void dumpEvent(int eventno, double weight, double xsec) {
    call_pyhepc( 1 );
    HepMC::GenEvent evt;
    evt.use_units(HepMC::Units::GEV, HepMC::Units::MM);
    hepevtio.fill_next_event(&evt);
    evt.set_event_number(eventno);
    HepMC::GenCrossSection xs;
    xs.set_cross_section(xsec);
    evt.set_cross_section( xs );
    evt.weights().push_back(weight);
    ofs << evt;
  }

private:


  std::ofstream ofs;
  HepMC::IO_HEPEVT hepevtio;


};


#endif
