//////////////////////////////////////////////////
//////////////////////////////////////////////////
// This source code is Copyright (2011) of      //
//  Jeppe R. Andersen and Jennifer M. Smillie   //
// and is distributed under the                 //
// Gnu Public License version 2                 //
// http://www.gnu.org/licenses/gpl-2.0.html     //
// You are allowed to distribute and alter the  //
// source under the conditions of the GPLv2     //
// as long as this copyright notice             //
// is unaltered and distributed with the source //
// Any use should comply with the               //
//             MCNET GUIDELINES                 //
//    for Event Generator Authors and Users     //
// as distributed with this source code         //
//////////////////////////////////////////////////
//////////////////////////////////////////////////
#include "Jets.h"
#include "TreeJets.h"
#include "MomentumDefJet.h"
//#include "Tree3jenu.h"

#define printalphasm 0

//const double beta0=11.-2.*5/3.;  // For nf=5

double CMultijet::MG(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb)
{
  double wt(0.);
  FREAL meout(0.),p12[4][4],p21[4][4];
  // FREAL meout2(0.);  
  // p12 = config where partons are ordered as in name
  // p21 = config where partons are reversed relative to name

  p12[0][0]=pb.e();    // forward parton
  p12[0][1]=pb.x();
  p12[0][2]=pb.y();
  p12[0][3]=pb.z();
  
  p12[1][0]=pa.e();   //  backward parton
  p12[1][1]=pa.x();
  p12[1][2]=pa.y();
  p12[1][3]=pa.z();
  
  p12[2][0]=(partons.back()).e();   // more forward jet
  p12[2][1]=(partons.back()).x();
  p12[2][2]=(partons.back()).y();
  p12[2][3]=(partons.back()).z();
  
  p12[3][0]=(partons.front()).e();  // more backward jet
  p12[3][1]=(partons.front()).x();
  p12[3][2]=(partons.front()).y();
  p12[3][3]=(partons.front()).z();

  for (int ii=0; ii<4; ii++) {
    p21[0][ii]=p12[1][ii];
    p21[1][ii]=p12[0][ii];
    p21[2][ii]=p12[3][ii];
    p21[3][ii]=p12[2][ii];
  }

  double idpartfact=1.;     //  Factor to take account of identical final states

  if (aptype==21) {// a=g
    if (bptype==21) {
      smatrixggtgg_(p12[0],&meout);
      idpartfact*=2.;
      //      std::cout<<"MG: "<<p12[0][3]<<"  "<<p12[1][3]<<"  "<<p12[2][3]<<"  "<<p12[3][3]<<" --> ";
      //      std::cout<<meout*idpartfact;
    }
    else if (bptype>0) 
      smatrixugtug_(p12[0],&meout);
    else 
      smatrixdxgtdxg_(p12[0],&meout);
  }
  else if (bptype==21) {  // b=g, a!=g
    if (aptype>0) 
      smatrixugtug_(p21[0],&meout);
    else 
      smatrixdxgtdxg_(p21[0],&meout);
  }
  else if (aptype==bptype) {
    idpartfact*=2.;
    if (aptype > 0)
      smatrixuutuu_(p21[0],&meout);
    else 
      smatrixdxdxtdxdx_(p12[0],&meout);
  }
  else if (aptype==-bptype) {
    if (aptype > 0)   // backward particle is q, forward qbar
      smatrixuuxtuux_(p21[0],&meout);
    else           // forward (b) particle is q, backward (a) qbar
      smatrixuuxtuux_(p12[0],&meout);
  }
  else if (aptype>0) {
    if (bptype>0) {
      smatrixustus_(p12[0],&meout);
      // smatrixustus_(p21[0],&meout2);
      // std::cerr<<"MG swap: "<<meout<<"  "<<meout2<<std::endl;
      //      std::cerr<<"MG: ("<<p12[0][0]<<",...)  ("<<p12[1][0]<<",...)  ("<<p12[2][0]<<",...)  ("<<p12[3][0]<<",...)\n";
    }
    else
      smatrixusxtusx_(p21[0],&meout);
  }
  else {
    if (bptype>0)
      smatrixusxtusx_(p12[0],&meout);
    else
      smatrixdxsxtdxsx_(p12[0],&meout);
  }

  wt=meout*idpartfact;
  //  std::cout<<" gives "<<wt<<"\n";

  double gs2=0.;
  if (flags.logcorrect==0) {
    gs2=4.*M_PI*Halphas(renscale);    // PTMAX set in weight() before MEt2 if req.
    double couplings=pow(gs2,2);
    wt*=couplings;
  } else {     // flags.logcorrect==1
    // wt*=pow(4.*M_PI,2)*Halphas(partons.front().perp()*flags.scalefactor)*Halphas(partons.back().perp()*flags.scalefactor);
    //    wt*=pow(4.*M_PI*Halphas(renscale),2)*(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI));
    wt*=pow(4.*M_PI,2)*NLLCorrect(partons.front().perp())*NLLCorrect(partons.back().perp());
#if printalphasm
    std::cout<<"MG2: "<<Halphas(partons.front().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partons.front().perp())<<"  "<<Halphas(partons.back().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partons.back().perp())<<"\n";
#endif
  }
  return wt;
}

double CMultijet::Generate2JNonFKLflavourconf_and_eval(double pmg[4][4])
{
  //    std::cout<<"Yep, 4j non FKL: ";
  double weight=1.;

  if (aptype==21) {// a=g 
    if (bptype==21) {
      
      static std::vector<sPMchoice> gg2Processes;
      if (gg2Processes.size()==0)  {
	  { // first gg process
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixggtuux_;
	    PMtemp.idpartmult=5;  //  5 flavours
	    
	    PMtemp.MomShuffleVector.push_back(0);
	    PMtemp.MomShuffleVector.push_back(1);
	    
	    gg2Processes.push_back(PMtemp);
	  }
      }
      weight*=PickMomAndEval2J(gg2Processes,pmg);
    }
    else if (bptype>0) {
      static std::vector<sPMchoice> gq2Processes;
      if (gq2Processes.size()==0)  {
	struct sPMchoice PMtemp;
	PMtemp.MGfunc=smatrixugtug_;
	PMtemp.idpartmult=1;
	PMtemp.MomShuffleVector.push_back(1);

	gq2Processes.push_back(PMtemp);
      }
      weight*=NoPickButEval2J(gq2Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> gqx2Processes;
      if (gqx2Processes.size()==0)  {
	struct sPMchoice PMtemp;
	PMtemp.MGfunc=smatrixdxgtdxg_;
	PMtemp.idpartmult=1;
	PMtemp.MomShuffleVector.push_back(1);

	gqx2Processes.push_back(PMtemp);
      }
      weight*=NoPickButEval2J(gqx2Processes,pmg);
    }
  }
  else if (bptype==21) {
    if (aptype>0) {
      static std::vector<sPMchoice> qg2Processes;
      if (qg2Processes.size()==0)  {
	struct sPMchoice PMtemp;
	PMtemp.MGfunc=smatrixugtug_;
	PMtemp.idpartmult=1;
	PMtemp.MomShuffleVector.push_back(3);

	qg2Processes.push_back(PMtemp);
      }
      weight*=NoPickButEval2J(qg2Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxg2Processes;
      if (qxg2Processes.size()==0)  {
	struct sPMchoice PMtemp;
	PMtemp.MGfunc=smatrixdxgtdxg_;
	PMtemp.idpartmult=1;
	PMtemp.MomShuffleVector.push_back(3);

	qxg2Processes.push_back(PMtemp);
      }
      weight*=NoPickButEval2J(qxg2Processes,pmg);
    }
  }
  //  If get down here, neither are gluons
  else if (aptype==bptype) {
    weight=0.;
  }
  else if (aptype==-bptype) {
    if (aptype>0) {
      static std::vector<sPMchoice> qqx2Processes;
      if (qqx2Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuuxtuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.MomShuffleVector.push_back(3);
	  
	  qqx2Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixssxtuux_;
	  PMtemp.idpartmult=4;  // 4 flavours
	  PMtemp.MomShuffleVector.push_back(2);
	  PMtemp.MomShuffleVector.push_back(3);
	  qqx2Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuuxtgg_;
	  PMtemp.idpartmult=2;  
	  PMtemp.MomShuffleVector.push_back(2);
	  qqx2Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom2J(qqx2Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxq2Processes;
      if (qxq2Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuuxtuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.MomShuffleVector.push_back(1);
	  qxq2Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixssxtuux_;
	  PMtemp.idpartmult=4;  // 4 flavours
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(1);
	  qxq2Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuuxtgg_;
	  PMtemp.idpartmult=2;  
	  PMtemp.MomShuffleVector.push_back(0);
	  qxq2Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom2J(qxq2Processes,pmg);
    }
  }
  else if (aptype>0) {
    if (bptype>0) {
      static std::vector<sPMchoice> qQ2Processes;
      if (qQ2Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustus_;
	  PMtemp.idpartmult=1;
	  PMtemp.MomShuffleVector.push_back(3);
	  
	  qQ2Processes.push_back(PMtemp);
	}
      }
      weight*=NoPickButEval2J(qQ2Processes,pmg);
      //      std::cout<<"Matching: "<<weight<<"\n";
    }
    else {
      static std::vector<sPMchoice> qQx2Processes;
      if (qQx2Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtusx_;
	  PMtemp.idpartmult=1;
	  PMtemp.MomShuffleVector.push_back(3);
	  
	  qQx2Processes.push_back(PMtemp);
	}
      }
      weight*=NoPickButEval2J(qQx2Processes,pmg);
      //      std::cout<<weight;
    }
  }
  else {
    if (bptype>0) {
      static std::vector<sPMchoice> qxQ2Processes;
      if (qxQ2Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtusx_;
	  PMtemp.idpartmult=1;
	  PMtemp.MomShuffleVector.push_back(1);
	  
	  qxQ2Processes.push_back(PMtemp);
	}
      }
      weight*=NoPickButEval2J(qxQ2Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxQx2Processes;
      if (qxQx2Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtdxsx_;
	  PMtemp.idpartmult=1;
	  PMtemp.MomShuffleVector.push_back(1);
	  
	  qxQx2Processes.push_back(PMtemp);
	}
      }
      weight*=NoPickButEval2J(qxQx2Processes,pmg);
    }
  }
  //  std::cout<<" gives "<<weight<<"  "<<flags.scale<<"\n";
  if (flags.logcorrect==0)
    weight*=pow(4.*M_PI*Halphas(renscale),2);
  else  {      // flags.logcorrect==1
    //    weight*=pow(4.*M_PI,2)*Halphas(partonmomenta.front().perp()*flags.scalefactor)*Halphas(partonmomenta.back().perp()*flags.scalefactor);
    //    weight*=pow(4.*M_PI*Halphas(renscale),2)*(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI));
    weight*=pow(4.*M_PI,2)*NLLCorrect(partonmomenta.front().perp())*NLLCorrect(partonmomenta.back().perp());
#if printalphasm
    std::cout<<"NonFKL2: "<<Halphas(partonmomenta.front().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partonmomenta.front().perp())<<"  "<<Halphas(partonmomenta.back().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partonmomenta.back().perp())<<"\n";
#endif
  }
  return weight;
}


double CMultijet::MG3(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb)
{
  double wt(0.);
  FREAL meout(0.),p12[5][4],p21[5][4];
  // p12 = config where partons are ordered as in name
  // p21 = config where partons are reversed relative to name

  p12[0][0]=pb.e();    // forward parton
  p12[0][1]=pb.x();
  p12[0][2]=pb.y();
  p12[0][3]=pb.z();
  
  p12[1][0]=pa.e();   //  backward parton
  p12[1][1]=pa.x();
  p12[1][2]=pa.y();
  p12[1][3]=pa.z();
  
  p12[2][0]=(partons.back()).e();   // most forward jet
  p12[2][1]=(partons.back()).x();
  p12[2][2]=(partons.back()).y();
  p12[2][3]=(partons.back()).z();
  
  p12[3][0]=(partons.at(1)).e();  // middle jet
  p12[3][1]=(partons.at(1)).x();
  p12[3][2]=(partons.at(1)).y();
  p12[3][3]=(partons.at(1)).z();

  p12[4][0]=(partons.front()).e();  // most backward jet
  p12[4][1]=(partons.front()).x();
  p12[4][2]=(partons.front()).y();
  p12[4][3]=(partons.front()).z();

  for (int ii=0; ii<4; ii++) {
    p21[0][ii]=p12[1][ii];
    p21[1][ii]=p12[0][ii];
    p21[2][ii]=p12[4][ii];
    p21[3][ii]=p12[3][ii];
    p21[4][ii]=p12[2][ii];
  }

  if (aptype==21) {// a=g
    if (bptype==21) {
      smatrixggtggg_(p12[0],&meout);
      meout*=6.;   
    }
    else {
      if (bptype>0)
	smatrixugtugg_(p12[0],&meout);
      else
	smatrixdxgtdxgg_(p12[0],&meout);
      meout*=2.;
    }
  }
  else if (bptype==21) {  // b=g, a!=g
    if (aptype>0) 
      smatrixugtugg_(p21[0],&meout);
    else 
      smatrixdxgtdxgg_(p21[0],&meout);
    meout*=2.;
  }
  else if (aptype==bptype) {
    if (aptype > 0) 
      smatrixuutugu_(p21[0],&meout);
    else 
      smatrixdxdxtdxgdx_(p12[0],&meout);
    meout*=2.;
  }
  else if (aptype==-bptype) {
    if (aptype > 0)   // backward particle is q, forward qbar
      smatrixuuxtugux_(p21[0],&meout);
    else             // forward (b) particle is q, backward (a) qbar
      smatrixuuxtugux_(p12[0],&meout);
  }
  else if (aptype>0) {
    if (bptype>0)
      smatrixustugs_(p12[0],&meout);
    else
      smatrixusxtugsx_(p21[0],&meout);
  }
  else {
    if (bptype>0)
      smatrixusxtugsx_(p12[0],&meout);
    else
      smatrixdxsxtdxgsx_(p12[0],&meout);
  }

  wt=meout;
  
  double gs2=0.;
  if (flags.logcorrect==0) {
    gs2=4.*M_PI*Halphas(renscale);    // =PTMAX set in weight() before MEt2 if req
    double couplings=pow(gs2,3);
    wt*=couplings;
  } else  {   // flags.logcorrect==1
    //    wt*=pow(4.*M_PI,3)*Halphas(partons.front().perp()*flags.scalefactor)*Halphas(partons.at(1).perp()*flags.scalefactor)*Halphas(partons.back().perp()*flags.scalefactor);
    //    wt*=pow(4.*M_PI*Halphas(renscale),3)*(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.at(1).perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI));
    wt*=pow(4.*M_PI,3)*NLLCorrect(partons.front().perp())*NLLCorrect(partons.at(1).perp())*NLLCorrect(partons.back().perp());
#if printalphasm
    std::cout<<"MG3: "<<Halphas(partons.front().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partons.front().perp())<<"  "<<Halphas(partons.at(1).perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.at(1).perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partons.at(1).perp())<<"  "<<Halphas(partons.back().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partons.back().perp())<<"\n";
#endif
  }

  return wt;
}

double CMultijet::Generate3JNonFKLflavourconf_and_eval(double pmg[5][4])
{
  //    std::cout<<"Yep, 3j non FKL: ";
  double weight=1.;

  if (aptype==21) {// a=g 
    if (bptype==21) {
      
      static std::vector<sPMchoice> gg3Processes;
      if (gg3Processes.size()==0)  {
	  { // first gg process
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixggtugux_;
	    PMtemp.idpartmult=5; // 5 flavours
	    
	    PMtemp.MomShuffleVector.push_back(0);
	    for (int ii=2;ii<7;ii++) 
	      PMtemp.MomShuffleVector.push_back(ii);
	    
	    gg3Processes.push_back(PMtemp);
	  }
      }
      weight*=PickMomAndEval3J(gg3Processes,pmg);
    }
    else if (bptype>0) {
      static std::vector<sPMchoice> gq3Processes;
      if (gq3Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtugg_;
	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(2);
	  PMtemp.MomShuffleVector.push_back(4);

	  gq3Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtuuux_;
	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(4);

	  gq3Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtussx_;
	  PMtemp.idpartmult=4;  // 4 flavours
	  PMtemp.MomShuffleVector.push_back(0);
	  for (int ii=2; ii<7; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  gq3Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom3J(gq3Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> gqx3Processes;
      if (gqx3Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxgg_;
	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(2);
	  PMtemp.MomShuffleVector.push_back(4);
	  
	  gqx3Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxdxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(4);
	  
	  gqx3Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxssx_;
	  PMtemp.idpartmult=4;  // 4 flavours
	  PMtemp.MomShuffleVector.push_back(0);
	  for (int ii=2; ii<7; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
	  gqx3Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom3J(gqx3Processes,pmg);
    }
  }
  else if (bptype==21) {
    if (aptype>0) {
      static std::vector<sPMchoice> qg3Processes;
      if (qg3Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtugg_;
	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);

	  qg3Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtuuux_;
	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(8);
	  PMtemp.MomShuffleVector.push_back(9);

	  qg3Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtussx_;
	  PMtemp.idpartmult=4;  // 4 flavours
	  PMtemp.MomShuffleVector.push_back(1);
	  for (int ii=7; ii<12; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  qg3Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom3J(qg3Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxg3Processes;
      if (qxg3Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxgg_;
	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);

	  qxg3Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxdxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(8);
	  PMtemp.MomShuffleVector.push_back(9);

	  qxg3Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxssx_;
	  PMtemp.idpartmult=4; // 4 flavours
	  PMtemp.MomShuffleVector.push_back(1);
	  for (int ii=7; ii<12; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  qxg3Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom3J(qxg3Processes,pmg);
    }
  }
  //  If get down here, neither are gluons
  else if (aptype==bptype) {
    if (aptype>0) {
      static std::vector<sPMchoice> qq3Processes;
      if (qq3Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuutugu_;
	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(2);
	  PMtemp.MomShuffleVector.push_back(3);

	  qq3Processes.push_back(PMtemp);
	}
      }
      weight*=PickMomAndEval3J(qq3Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxqx3Processes;
      if (qxqx3Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxdxtdxgdx_;
	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(2);
	  PMtemp.MomShuffleVector.push_back(3);

	  qxqx3Processes.push_back(PMtemp);
	}
      }
      weight*=PickMomAndEval3J(qxqx3Processes,pmg);
    }
  }
  else if (aptype==-bptype) {
    if (aptype>0) {
      static std::vector<sPMchoice> qqx3Processes;
      if (qqx3Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuuxtugux_;
  	  PMtemp.idpartmult=1;

	  for (int ii=7; ii<12; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qqx3Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixssxtugux_;
  	  PMtemp.idpartmult=4;  // 4 flavours
  	  PMtemp.MomShuffleVector.push_back(1);
	  for (int ii=7; ii<12; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
  	  qqx3Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixddxtggg_;
  	  PMtemp.idpartmult=6;  
  	  PMtemp.MomShuffleVector.push_back(1);
  	  qqx3Processes.push_back(PMtemp);
  	}
      }
      weight*=PickAndEvalProcAndMom3J(qqx3Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxq3Processes;
      if (qxq3Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuuxtugux_;
  	  PMtemp.idpartmult=1;
	  for (int ii=2; ii<7; ii++) 
	    PMtemp.MomShuffleVector.push_back(ii);
  	  qxq3Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixssxtugux_;
  	  PMtemp.idpartmult=4;  // 4 flavours
  	  PMtemp.MomShuffleVector.push_back(0);
	  for (int ii=2; ii<7; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
  	  qxq3Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixddxtggg_;
  	  PMtemp.idpartmult=6;  
  	  PMtemp.MomShuffleVector.push_back(0);
  	  qxq3Processes.push_back(PMtemp);
  	}
      }
      weight*=PickAndEvalProcAndMom3J(qxq3Processes,pmg);
    }
  }
  else if (aptype>0) {
    if (bptype>0) {
      static std::vector<sPMchoice> qQ3Processes;
      if (qQ3Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixustugs_;
  	  PMtemp.idpartmult=1;
	  for (int ii=2; ii<7; ii++) 
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qQ3Processes.push_back(PMtemp);
  	}
      }
      weight*=PickMomAndEval3J(qQ3Processes,pmg);
      //      std::cout<<"Matching: "<<weight<<"\n";
    }
    else {
      static std::vector<sPMchoice> qQx3Processes;
      if (qQx3Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixusxtugsx_;
  	  PMtemp.idpartmult=1;
	  for (int ii=7; ii<12; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qQx3Processes.push_back(PMtemp);
  	}
      }
      weight*=PickMomAndEval3J(qQx3Processes,pmg);
    }
  }
  else {
    if (bptype>0) {
      static std::vector<sPMchoice> qxQ3Processes;
      if (qxQ3Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixusxtugsx_;
  	  PMtemp.idpartmult=1;
	  for (int ii=2; ii<7; ii++) 
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qxQ3Processes.push_back(PMtemp);
  	}
      }
      weight*=PickMomAndEval3J(qxQ3Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxQx3Processes;
      if (qxQx3Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixdxsxtdxgsx_;
  	  PMtemp.idpartmult=1;
	  for (int ii=2; ii<7; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qxQx3Processes.push_back(PMtemp);
  	}
      }
      weight*=PickMomAndEval3J(qxQx3Processes,pmg);
    }
  }

  if (flags.logcorrect==0) 
    weight*=pow(4.*M_PI*Halphas(renscale),3);
  else {    // flags.logcorrect==1
    //    weight*=pow(4.*M_PI,3)*Halphas(partonmomenta.front().perp()*flags.scalefactor)*Halphas(partonmomenta.at(1).perp()*flags.scalefactor)*Halphas(partonmomenta.back().perp()*flags.scalefactor);
    //    weight*=pow(4.*M_PI*Halphas(renscale),3)*(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI));
    weight*=pow(4.*M_PI,3)*NLLCorrect(partonmomenta.front().perp())*NLLCorrect(partonmomenta.at(1).perp())*NLLCorrect(partonmomenta.back().perp());
#if printalphasm
    std::cout<<"NonFKL3: "<<Halphas(partonmomenta.front().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partonmomenta.front().perp())<<"  "<<Halphas(partonmomenta.at(1).perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partonmomenta.at(1).perp())<<"  "<<Halphas(partonmomenta.back().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partonmomenta.back().perp())<<"\n";
#endif
  }
  
  return weight;
}


double CMultijet::MG4(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb)
{
  double wt(0.);
  FREAL meout(0.),p12[6][4],p21[6][4];
  // p12 = config where partons are ordered as in name
  // p21 = config where partons are reversed relative to name

  p12[0][0]=pb.e();    // forward parton
  p12[0][1]=pb.x();
  p12[0][2]=pb.y();
  p12[0][3]=pb.z();
  
  p12[1][0]=pa.e();   //  backward parton
  p12[1][1]=pa.x();
  p12[1][2]=pa.y();
  p12[1][3]=pa.z();
  
  p12[2][0]=(partons.back()).e();   // most forward jet
  p12[2][1]=(partons.back()).x();
  p12[2][2]=(partons.back()).y();
  p12[2][3]=(partons.back()).z();
  
  p12[3][0]=(partons.at(2)).e();  // middle forward jet
  p12[3][1]=(partons.at(2)).x();
  p12[3][2]=(partons.at(2)).y();
  p12[3][3]=(partons.at(2)).z();

  p12[4][0]=(partons.at(1)).e();  // middle backward jet
  p12[4][1]=(partons.at(1)).x();
  p12[4][2]=(partons.at(1)).y();
  p12[4][3]=(partons.at(1)).z();

  p12[5][0]=(partons.front()).e();  // most backward jet
  p12[5][1]=(partons.front()).x();
  p12[5][2]=(partons.front()).y();
  p12[5][3]=(partons.front()).z();

  for (int ii=0; ii<4; ii++) {
    p21[0][ii]=p12[1][ii];
    p21[1][ii]=p12[0][ii];
    p21[2][ii]=p12[5][ii];
    p21[3][ii]=p12[4][ii];
    p21[4][ii]=p12[3][ii];
    p21[5][ii]=p12[2][ii];
  }

  if (aptype==21) {// a=g
    if (bptype==21) {
      smatrixggtgggg_(p12[0],&meout);
      meout*=24.;
    }
    else if (bptype>0) {
      smatrixugtuggg_(p12[0],&meout);
      meout*=6.;
    }
    else {
      smatrixdxgtdxggg_(p12[0],&meout);
      meout*=6.;
    }
  }
  else if (bptype==21) {  // b=g, a!=g
    if (aptype>0) 
      smatrixugtuggg_(p21[0],&meout);
    else 
      smatrixdxgtdxggg_(p21[0],&meout);
    meout*=6.;
  }
  else if (aptype==bptype) {
    if (aptype > 0)
      smatrixuutuggu_(p12[0],&meout);
    else 
      smatrixdxdxtdxggdx_(p21[0],&meout);
    meout*=4.;
    }
  else if (aptype==-bptype) {
    if (aptype > 0)   // backward particle is q, forward qbar
      smatrixuuxtuggux_(p21[0],&meout);
    else             // forward (b) particle is q, backward (a) qbar
      smatrixuuxtuggux_(p21[0],&meout);
    meout*=2.;
  }
  else if (aptype>0) {
    if (bptype>0)
      smatrixustuggs_(p12[0],&meout);
    else
      smatrixusxtuggsx_(p21[0],&meout);
    meout*=2.;
  }
  else {
    if (bptype>0)
      smatrixusxtuggsx_(p12[0],&meout);
    else
      smatrixdxsxtdxggsx_(p12[0],&meout);
    meout*=2.;
  }

  wt=meout;
  
  double gs2=0.;
  if (flags.logcorrect==0) {
    gs2=4.*M_PI*Halphas(renscale);    // =PTMAX set in weight() before MEt2 if req
    double couplings=pow(gs2,4);
    wt*=couplings;
  } else {  // flags.logcorrect==1
    //    wt*=pow(4.*M_PI,4)*Halphas(partons.front().perp()*flags.scalefactor)*Halphas(partons.at(1).perp())*Halphas(partons.at(2).perp()*flags.scalefactor)*Halphas(partons.back().perp()*flags.scalefactor);
    //    wt*=pow(4.*M_PI*Halphas(renscale),4)*(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.at(1).perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.at(2).perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI));
    wt*=pow(4.*M_PI,4)*NLLCorrect(partons.front().perp())*NLLCorrect(partons.at(1).perp())*NLLCorrect(partons.at(2).perp())*NLLCorrect(partons.back().perp());
#if printalphasm
    std::cout<<"MG4: "<<Halphas(partons.front().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partons.front().perp())<<"  "<<Halphas(partons.at(1).perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.at(1).perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partons.at(1).perp())<<"  "<<Halphas(partons.at(2).perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.at(2).perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partons.at(2).perp())<<"  "<<Halphas(partons.back().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partons.back().perp())<<"\n";
#endif
  }

  return wt;
}

double CMultijet::Generate4JNonFKLflavourconf_and_eval(double pmg[6][4])
{
  //    std::cout<<"Yep, 4j non FKL: ";
  double weight=1.;

  if (aptype==21) {// a=g 
    if (bptype==21) {
      
      static std::vector<sPMchoice> gg4Processes;
      if (gg4Processes.size()==0)  {
	{ // first gg process
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixggtuggux_;
	  PMtemp.idpartmult=10; // 5 flavours * 2 id parts
	  
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(14);
	  PMtemp.MomShuffleVector.push_back(16);
	  PMtemp.MomShuffleVector.push_back(17);
	  PMtemp.MomShuffleVector.push_back(20);
	  PMtemp.MomShuffleVector.push_back(22);
	  PMtemp.MomShuffleVector.push_back(23);
	  
	  gg4Processes.push_back(PMtemp);
	}
	{ // second gg process
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixggtuuuxux_;
	  PMtemp.idpartmult=20; // 5 flavours * 4 id parts
	  
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(2);
	  PMtemp.MomShuffleVector.push_back(4);
	  PMtemp.MomShuffleVector.push_back(8);
	  PMtemp.MomShuffleVector.push_back(10);
	  PMtemp.MomShuffleVector.push_back(12);
	  
	  gg4Processes.push_back(PMtemp);
	}
	{ // third gg process
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixggtuuxssx_;
	  PMtemp.idpartmult=10; // 5*4/2 flavours
	  
	  for (int ii=0;ii<24;ii++) 
	    PMtemp.MomShuffleVector.push_back(ii);
	  
	  gg4Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4J(gg4Processes,pmg);
    }
    else if (bptype>0) {
      static std::vector<sPMchoice> gq4Processes;
      if (gq4Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtuggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(18);

	  gq4Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtuuuxg_;
	  PMtemp.idpartmult=2;  
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  gq4Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtussxg_;
	  PMtemp.idpartmult=4;  // 4 flavours
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  gq4Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4J(gq4Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> gqx4Processes;
      if (gqx4Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(18);
	  
	  gqx4Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxdxdg_;
	  PMtemp.idpartmult=2;
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
	  gqx4Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxssxg_;
	  PMtemp.idpartmult=4;  // 4 flavours
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
	  gqx4Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4J(gqx4Processes,pmg);
    }
  }
  else if (bptype==21) {
    if (aptype>0) {
      static std::vector<sPMchoice> qg4Processes;
      if (qg4Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtuggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(42);

	  qg4Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtussxg_;
	  PMtemp.idpartmult=4;  // 4 flavours
	  for (int ii=24; ii<48; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  qg4Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtuuuxg_;
	  PMtemp.idpartmult=2;
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  qg4Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4J(qg4Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxg4Processes;
      if (qxg4Processes.size()==0)  {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(42);

	  qxg4Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxssxg_;
	  PMtemp.idpartmult=4; // 4 flavours
	  for (int ii=24; ii<48; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  qxg4Processes.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtdxdxdg_;
	  PMtemp.idpartmult=2; 
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  qxg4Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4J(qxg4Processes,pmg);
    }
  }
  //  If get down here, neither are gluons
  else if (aptype==bptype) {
    if (aptype>0) {
      static std::vector<sPMchoice> qq4Processes;
      if (qq4Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuutuggu_;
  	  PMtemp.idpartmult=4;
  	  PMtemp.MomShuffleVector.push_back(1);
  	  PMtemp.MomShuffleVector.push_back(3);
  	  PMtemp.MomShuffleVector.push_back(8);
  	  PMtemp.MomShuffleVector.push_back(10);
  	  PMtemp.MomShuffleVector.push_back(15);

  	  qq4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuutuuuux_;
  	  PMtemp.idpartmult=6;
  	  PMtemp.MomShuffleVector.push_back(0);
  	  PMtemp.MomShuffleVector.push_back(1);
  	  PMtemp.MomShuffleVector.push_back(3);
  	  PMtemp.MomShuffleVector.push_back(9);

  	  qq4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuutuussx_;
  	  PMtemp.idpartmult=8;   // 4 flavours * 2 id part
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

  	  qq4Processes.push_back(PMtemp);
  	}
      }
      weight*=PickAndEvalProcAndMom4J(qq4Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxqx4Processes;
      if (qxqx4Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixdxdxtdxggdx_;
  	  PMtemp.idpartmult=4;
  	  PMtemp.MomShuffleVector.push_back(1);
  	  PMtemp.MomShuffleVector.push_back(3);
  	  PMtemp.MomShuffleVector.push_back(8);
  	  PMtemp.MomShuffleVector.push_back(10);
  	  PMtemp.MomShuffleVector.push_back(15);

  	  qxqx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixdxdxtdxdxdxd_;
  	  PMtemp.idpartmult=6;
  	  PMtemp.MomShuffleVector.push_back(0);
  	  PMtemp.MomShuffleVector.push_back(1);
  	  PMtemp.MomShuffleVector.push_back(3);
  	  PMtemp.MomShuffleVector.push_back(9);

  	  qxqx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixdxdxtdxdxssx_;
  	  PMtemp.idpartmult=8;  // 4 flavours * 2 id part
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

  	  qxqx4Processes.push_back(PMtemp);
  	}
      }
      weight*=PickAndEvalProcAndMom4J(qxqx4Processes,pmg);
    }
  }
  else if (aptype==-bptype) {
    if (aptype>0) {
      static std::vector<sPMchoice> qqx4Processes;
      if (qqx4Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuuxtuggux_;
  	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(38);
	  PMtemp.MomShuffleVector.push_back(40);
	  PMtemp.MomShuffleVector.push_back(41);
	  PMtemp.MomShuffleVector.push_back(44);
	  PMtemp.MomShuffleVector.push_back(46);
	  PMtemp.MomShuffleVector.push_back(47);

  	  qqx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuuxtuuuxux_;
  	  PMtemp.idpartmult=4; 
	  for (int ii=12; ii<15; ii++)
	    PMtemp.MomShuffleVector.push_back(ii*2);
	  for (int ii=16; ii<19; ii++)
	    PMtemp.MomShuffleVector.push_back(ii*2);
	  
  	  qqx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuuxtussxux_;
  	  PMtemp.idpartmult=4;  
	  for (int ii=24; ii<48; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qqx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixssxtuggux_;
  	  PMtemp.idpartmult=8;  // 4 flavours, 2 id part
	  PMtemp.MomShuffleVector.push_back(24);
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(38);
	  PMtemp.MomShuffleVector.push_back(40);
	  PMtemp.MomShuffleVector.push_back(41);
	  PMtemp.MomShuffleVector.push_back(44);
	  PMtemp.MomShuffleVector.push_back(46);
	  PMtemp.MomShuffleVector.push_back(47);
  	  qqx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixssxtuuuxux_;
  	  PMtemp.idpartmult=16;  // 4 flavours * 4 id part
  	  for (int ii=12; ii<15; ii++)
  	    PMtemp.MomShuffleVector.push_back(2*ii);
  	  for (int ii=16; ii<19; ii++)
  	    PMtemp.MomShuffleVector.push_back(2*ii);
  	  qqx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixssxtuddxux_;
  	  PMtemp.idpartmult=6;  // 4*3/2 flavours
  	  for (int ii=24; ii<48; ii++)
  	    PMtemp.MomShuffleVector.push_back(ii);
  	  qqx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixddxtgggg_;
  	  PMtemp.idpartmult=24;  
  	  PMtemp.MomShuffleVector.push_back(24);
  	  qqx4Processes.push_back(PMtemp);
  	}
      }
      weight*=PickAndEvalProcAndMom4J(qqx4Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxq4Processes;
      if (qxq4Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuuxtuggux_;
  	  PMtemp.idpartmult=2; 
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(14);
	  PMtemp.MomShuffleVector.push_back(16);
	  PMtemp.MomShuffleVector.push_back(17);
	  PMtemp.MomShuffleVector.push_back(20);
	  PMtemp.MomShuffleVector.push_back(22);
	  PMtemp.MomShuffleVector.push_back(23);
  	  qxq4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuuxtuuuxux_;
  	  PMtemp.idpartmult=4; 
  	  for (int ii=0; ii<3; ii++) 
  	    PMtemp.MomShuffleVector.push_back(ii*2);
  	  for (int ii=4; ii<7; ii++) 
  	    PMtemp.MomShuffleVector.push_back(ii*2);
  	  qxq4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixuuxtussxux_;
  	  PMtemp.idpartmult=4;  // 4 flavours
  	  for (int ii=0; ii<24; ii++) 
  	    PMtemp.MomShuffleVector.push_back(ii);
  	  qxq4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixssxtuggux_;
  	  PMtemp.idpartmult=8;  // 4 flavours * 2 id part
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(14);
	  PMtemp.MomShuffleVector.push_back(16);
	  PMtemp.MomShuffleVector.push_back(17);
	  PMtemp.MomShuffleVector.push_back(20);
	  PMtemp.MomShuffleVector.push_back(22);
	  PMtemp.MomShuffleVector.push_back(23);

  	  qxq4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixssxtuuuxux_;
  	  PMtemp.idpartmult=16;  // 4 flavours * 4 id part
  	  for (int ii=0; ii<3; ii++) 
  	    PMtemp.MomShuffleVector.push_back(ii*2);
  	  for (int ii=4; ii<7; ii++) 
  	    PMtemp.MomShuffleVector.push_back(ii*2);
  	  qxq4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixssxtuddxux_;
  	  PMtemp.idpartmult=6;  // 4*3/2 flavours
  	  for (int ii=0; ii<24; ii++)
  	    PMtemp.MomShuffleVector.push_back(ii);
  	  qxq4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixddxtgggg_;
  	  PMtemp.idpartmult=24;  
  	  PMtemp.MomShuffleVector.push_back(0);
  	  qxq4Processes.push_back(PMtemp);
  	}
      }
      weight*=PickAndEvalProcAndMom4J(qxq4Processes,pmg);
    }
  }
  else if (aptype>0) {
    if (bptype>0) {
      static std::vector<sPMchoice> qQ4Processes;
      if (qQ4Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixustuggs_;
  	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(38);
	  PMtemp.MomShuffleVector.push_back(40);
	  PMtemp.MomShuffleVector.push_back(41);
	  PMtemp.MomShuffleVector.push_back(44);
	  PMtemp.MomShuffleVector.push_back(46);
	  PMtemp.MomShuffleVector.push_back(47);
	  
  	  qQ4Processes.push_back(PMtemp);
  	}
	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixustuuuxs_;
  	  PMtemp.idpartmult=2;
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

  	  qQ4Processes.push_back(PMtemp);
	}
	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixustuuuxs_;
  	  PMtemp.idpartmult=2;
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

  	  qQ4Processes.push_back(PMtemp);
	}
	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixustuddxs_;
  	  PMtemp.idpartmult=3;   // 3 flavours
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

  	  qQ4Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4J(qQ4Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qQx4Processes;
      if (qQx4Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixusxtuggsx_;
  	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(38);
	  PMtemp.MomShuffleVector.push_back(40);
	  PMtemp.MomShuffleVector.push_back(41);
	  PMtemp.MomShuffleVector.push_back(44);
	  PMtemp.MomShuffleVector.push_back(46);
	  PMtemp.MomShuffleVector.push_back(47);
	  
  	  qQx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixusxtuuuxsx_;
  	  PMtemp.idpartmult=2;
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qQx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixusxtsxsxsu_;
  	  PMtemp.idpartmult=2;
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qQx4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixusxtuddxsx_;
  	  PMtemp.idpartmult=3;  // 3 flavours
	  for (int ii=24; ii<48; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qQx4Processes.push_back(PMtemp);
  	}
      }
      weight*=PickAndEvalProcAndMom4J(qQx4Processes,pmg);
    }
  }
  else {
    if (bptype>0) {
      static std::vector<sPMchoice> qxQ4Processes;
      if (qxQ4Processes.size()==0)  {
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixusxtuggsx_;
  	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(14);
	  PMtemp.MomShuffleVector.push_back(16);
	  PMtemp.MomShuffleVector.push_back(17);
	  PMtemp.MomShuffleVector.push_back(20);
	  PMtemp.MomShuffleVector.push_back(22);
	  PMtemp.MomShuffleVector.push_back(23);
	  
  	  qxQ4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixusxtuuuxsx_;
  	  PMtemp.idpartmult=2;
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qxQ4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixusxtsxsxsu_;
  	  PMtemp.idpartmult=2;
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qxQ4Processes.push_back(PMtemp);
  	}
  	{
  	  struct sPMchoice PMtemp;
  	  PMtemp.MGfunc=smatrixusxtuddxsx_;
  	  PMtemp.idpartmult=3;  // 3 flavours
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  
  	  qxQ4Processes.push_back(PMtemp);
  	}
      }
      weight*=PickAndEvalProcAndMom4J(qxQ4Processes,pmg);
    }
    else {
      static std::vector<sPMchoice> qxQx4Processes;
      if (qxQx4Processes.size()==0)  {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtdxggsx_;
    	  PMtemp.idpartmult=2;
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(38);
	  PMtemp.MomShuffleVector.push_back(40);
	  PMtemp.MomShuffleVector.push_back(41);
	  PMtemp.MomShuffleVector.push_back(44);
	  PMtemp.MomShuffleVector.push_back(46);
	  PMtemp.MomShuffleVector.push_back(47);
	  
    	  qxQx4Processes.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtdxdxdsx_;
    	  PMtemp.idpartmult=2;
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  qxQx4Processes.push_back(PMtemp);
	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtdxdxdsx_;
    	  PMtemp.idpartmult=2;
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  qxQx4Processes.push_back(PMtemp);
	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtdxuuxsx_;
    	  PMtemp.idpartmult=3; // 3 flavours
	  for (int ii=24; ii<48; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);

	  qxQx4Processes.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4J(qxQx4Processes,pmg);
    }
  }
  if (flags.logcorrect==0)
    weight*=pow(4.*M_PI*Halphas(renscale),4);
  else {   // flags.logcorrect==1
    //    weight*=pow(4.*M_PI,4)*Halphas(partonmomenta.front().perp()*flags.scalefactor)*Halphas(partonmomenta.at(1).perp())*Halphas(partonmomenta.at(2).perp()*flags.scalefactor)*Halphas(partonmomenta.back().perp()*flags.scalefactor);
    //    weight*=pow(4.*M_PI*Halphas(renscale),4)*(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(2).perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI));
    weight*=pow(4.*M_PI,4)*NLLCorrect(partonmomenta.front().perp())*NLLCorrect(partonmomenta.at(1).perp())*NLLCorrect(partonmomenta.at(2).perp())*NLLCorrect(partonmomenta.back().perp());
#if printalphasm
    std::cout<<"NonFKL4: "<<Halphas(partonmomenta.front().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partonmomenta.front().perp())<<"  "<<Halphas(partonmomenta.at(1).perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partonmomenta.at(1).perp())<<"  "<<Halphas(partonmomenta.at(2).perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(2).perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partonmomenta.at(2).perp())<<"  "<<Halphas(partonmomenta.back().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))<<"  "<<NLLCorrect(partonmomenta.back().perp())<<"\n";
#endif
  }
  //  std::cout<<"Matching: "<<weight<<"  "<<aptype<<"  "<<bptype<<std::endl;
  return weight;
}
